/*
   arguments.c
   
   THIS SOFTWARE IS IN PUBLIC DOMAIN
     There is no limitation on any kind of usage of this code. 
     You use it because it is yours.
     
     Originally written and maintained by Matlib (matlib@matlibhax.com),
     any comments or suggestions are welcome.
     
   Change log:
        1  2007-03-05  Matlib
        2  2009-02-19  Matlib (minor corrections)
        3  2011-02-15  Matlib (union, consts, attr and -O2 fix
                               + arguments_print_par 
                               + arguments_clr_occurrence)
        4  2011-03-23  Matlib (corrections to argument_u)
*/



#include "arguments.h"


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>
#include <limits.h>
#include <math.h>





#ifndef __USE_GNU
	static char const *program_invocation_short_name;
#else
	static extern char *program_invocation_short_name;
#endif

void arguments_error_freeform (char const *format, ...)
{
	va_list ap;

	va_start (ap, format);
	vfprintf (stderr, format, ap);
	va_end (ap);
	fprintf (stderr, "\nUse %s --help to see the list of available options.\n",
		 program_invocation_short_name);
	exit (EXIT_FAILURE);
}

void arguments_error (struct Argument const *const arg, 
	 char const *const string)
{
	fprintf (stderr, "Option ");
	arguments_print_arg (arg, stderr);
	arguments_error_freeform (" %s", string);
}

void arguments_print_arg (struct Argument const *const arg, 
	 FILE *const f)
{
	if ((arg -> lng != NULL) && (arg -> shrt) != 0)
		fprintf (f, "--%s (-%c)", arg -> lng, arg -> shrt);
	else if (arg -> lng != NULL)
		fprintf (f, "--%s", arg -> lng);
	else
		fprintf (f, "-%c", arg -> shrt);
}







ARGUMENT_CALLBACK (argument_bool)
{
	arg -> value_u ++;
}

ARGUMENT_CALLBACK (argument_string)
{
	arg -> value_str = parameter;
}

ARGUMENT_CALLBACK (argument_vs)
{
	if ((arg -> value_vs = strdup (parameter)) == NULL)
	{
		fprintf (stderr, "%s: libarguments: malloc: %s\n", 
			 program_invocation_short_name, strerror (errno));
		exit (EXIT_FAILURE);
	}
}


static const char *INT_RANGE_STR = 
	 "Integer value outside of allowed range for option ";

ARGUMENT_CALLBACK (argument_int)
{
	char *tail;
	arg -> value_int = strtol (parameter, &tail, 0);
	if (*tail != 0)
		arguments_error (arg, "expects integer as parameter.");
	if (errno == ERANGE)
	{
		fprintf (stderr, INT_RANGE_STR);
		arguments_print_arg (arg, stderr);
		arguments_error_freeform (".");
	}
}

ARGUMENT_CALLBACK (argument_u)
{
	char *tail;
	long long v;
	v = strtoll (parameter, &tail, 0);
	if ((*tail != 0) || (v < 0))
		arguments_error (arg, "expects positive integer as parameter.");
	if (v > ULONG_MAX)
	{
		fprintf (stderr, INT_RANGE_STR);
		arguments_print_arg (arg, stderr);
		arguments_error_freeform (".");
	}
	arg -> value_u = v;
}



ARGUMENT_CALLBACK (argument_fp)
{
	char *tail;
	errno = 0;
	#if _LP64 == 1
		double v;
		v = strtod (parameter, &tail);
	#else
		float v;
		v = strtof (parameter, &tail);
	#endif
	if (*tail)
		arguments_error (arg, "expects a number as parameter.");
	if (errno || !isfinite (v))
	{
		fprintf (stderr, "Number value is outside of allowed range for option ");
		arguments_print_arg (arg, stderr);
		arguments_error_freeform (".");
	}
	arg -> value_fp = v;
}






void arguments_version (const char *program_name, const char *program_version)
{
	if (program_name != NULL)
	{
		printf (program_name);
		if (program_version != NULL)
			printf (" version %s", program_version);
		putchar ('\n');
	}
}

void arguments_help_print_arg (const struct Argument *arg, int *f)
{
	if (arg -> shrt != 0)
	{
		printf ("  -%c", (*arg).shrt);
		if (arg -> shrt == 'h')
			*f |= 1; // flags = 1 (-h) | 2 (--help) | 4 (--version)
	}
	else
		printf ("    ");
	if (arg -> lng != NULL)
	{
		printf ("  --%s", (*arg).lng);
		if (strcmp (arg -> lng, "help") == 0)
			*f |= 2; // flags again, see note above
		else if (strcmp (arg -> lng, "version") == 0)
			*f |= 4;
	}
	if ((arg -> flags & ARGUMENT_F_PAR) != 0)
		printf (" <par>");
	putchar ('\n');
}

void arguments_help (char const *const *const argv, 
	 struct Argument const *const arg, 
	 char const *const program_name, char const *const program_version, 
	 char const *const usage, char const *const description)
{
	int i, f, nl;
	arguments_version (program_name, program_version);

	f = 0;
	printf ("Usage: %s [OPTION]... ", argv [0]);
	if (usage != NULL)
		printf (usage);
	putchar ('\n');
	if (description != NULL)
		printf (description);
	putchar ('\n');
	nl = 0;
	if (arg != NULL)
		for (i = 0; arg [i].cb != NULL; i ++)
			if ((arg [i].flags & ARGUMENT_F_OBL) != 0)
			{
				if (nl == 0)
				{
					putchar ('\n');
					nl ++;
				}
				arguments_help_print_arg (&(arg [i]), &f);
			}
	printf ("\nOptional arguments:\n\n");
	if (arg != NULL)
		for (i = 0; arg [i].cb != NULL; i ++)
			if ((arg [i].flags & ARGUMENT_F_OBL) == 0)
				arguments_help_print_arg (&(arg [i]), &f);

	/* print --help and --version if not redefined */
	if (f != 7)
	{
		putchar ('\n');
		printf ((f & 1) == 0 ? "  -h" : "    ");
		if ((f & 2) == 0)
			printf ("  --help");
		putchar ('\n');
		if ((f & 4) == 0)
			printf ("      --version\n");
	}

	putchar ('\n');   
	exit (EXIT_SUCCESS);
}





static void arguments_mark_option (char const *const *const argv, 
   struct Argument *const arg)
{
	if ((arg -> flags & (ARGUMENT_F_OCC | ARGUMENT_F_ONCE)) 
		 == (ARGUMENT_F_OCC | ARGUMENT_F_ONCE))
	arguments_error (arg, "specified more than once.");
	arg -> flags |= ARGUMENT_F_OCC;
}

int arguments_parse (char const *program_name, char const *program_version, 
	 char const *usage, char const *short_description, int const argc, 
	 char const *const *const argv, struct Argument *const arg)
{
	int a, i, pos;
	#ifndef __USE_GNU
		if (program_invocation_short_name == NULL)
			program_invocation_short_name = argv [0];
	#endif
	for (a = 1; (a < argc) && (strlen (argv [a]) > 1) && 
		 (argv [a][0] == '-'); a ++)
	{
		i = 0;
		if (argv [a][1] == '-')     // --long options
		{
			if (argv [a][2] == 0)     // stop processing at --
			{
				a ++;
				break;
			}
			for (pos = 0; (argv [a][pos] != 0) && (argv [a][pos] != '='); pos ++);
lloop:
			if ((arg == NULL) || (arg [i].cb == NULL))
			{
				if (strcmp (argv [a], "--version") == 0)
				{
					arguments_version (program_name, program_version);
					exit (EXIT_SUCCESS);
				}
				if (strcmp (argv [a], "--help") == 0)
					arguments_help (argv, arg, program_name, program_version, usage, 
						 short_description);
					arguments_error_freeform ("Unknown option: %s.", argv [a]);
			}
			if ((arg [i].lng == NULL) 
				 || (strncmp (arg [i].lng, argv [a] + 2, pos - 2  ) != 0))
			{
				i ++;
				goto lloop;
			}
			arguments_mark_option (argv, &(arg [i]));
			if (argv [a][pos] == '=')
			{
				pos ++;
				if ((arg [i].flags & ARGUMENT_F_PAR) == 0)
					arguments_error (&(arg [i]), "does not accept any parameters.");
			}
			if ((arg [i].flags & ARGUMENT_F_PAR) == 0)
				arg [i].cb (argc, argv, arg + i, NULL);
		}
		else                        // -s hort options
		{
			pos = 1;
sloop:
			if ((arg == NULL) || (arg [i].cb == NULL))
			{
				if (argv [a][pos] == 'h')
					arguments_help (argv, arg, program_name, program_version, usage, 
						 short_description);
				else
					arguments_error_freeform ("Unknown option: -%c.", argv [a][pos]);
			}
			if (arg [i].shrt != argv [a][pos])
			{
				i ++;
				goto sloop;
			}
			arguments_mark_option (argv, arg + i);
			pos ++;
			if ((arg [i].flags & ARGUMENT_F_PAR) == 0)
			{
				arg [i].cb (argc, argv, arg + i, NULL);
				i = 0;
				if (argv [a][pos] != 0)
					goto sloop;
				continue;
			}
		}

		if ((arg [i].flags & ARGUMENT_F_PAR) != 0)
		{
			if (argv [a][pos] == 0)
			{
				a ++;
				pos = 0;
				if (a >= argc)
					arguments_error (&(arg [i]), "requires parameter.");
			}
			arg [i].cb (argc, argv, arg + i, argv [a] + pos);
		}
	}


	// Checking which obligatory arguments were not found
	if (arg != NULL)
		for (i = 0; arg [i].cb != NULL; i ++)
			if (((arg [i].flags & ARGUMENT_F_OBL) != 0) 
				 && ((arg [i].flags & ARGUMENT_F_OCC) == 0))
				arguments_error (&(arg [i]), "must be specified.");

	return a;
}



void arguments_clr_occurrence (struct Argument *const arg)
{
	unsigned i;
	for (i = 0; arg [i].cb != NULL; i ++)
		if ((arg [i].flags & ARGUMENT_F_RET) == 0)
			arg [i].flags &= ~ARGUMENT_F_OCC;
}

